//
// frspider.cpp
//
// Copyright (c) 2003 Virtual Terrain Project
// Free for all uses, see license.txt for details.
//

#include "TripDub.h"
#include <map>
#include <string>
#include <vector>
using namespace std;

//-------------------------------------------------------------------------

void ExtractLine(const String2 &in, int offset, string &out)
{
	out = "";

	char c;
	while ((c = in[offset]) != '\n')
	{
		out += c;
		offset++;
	}
}

void SkipToNextLine(const String2 &str, int &offset)
{
	while (str[offset] != '\n')
		offset++;
	offset++;
}

#define PT_UNKNOWN		0
#define PT_SUSPENDED	1
#define PT_REALPERSON	2
#define PT_PLACE		3
#define PT_SUBSTANCE	4
#define PT_OTHER_FAKE	5

class Person
{
public:
	Person();

	bool QueryInfo();
	bool QueryFriends();
	int NumFriends() { return m_friends.size(); }

	int			m_id;
	String2		m_nick;
	String2		m_address;
	std::vector<int> m_friends;
	int			m_hops;
	int			m_type;
};

Person::Person()
{
	m_id = 0;
	m_hops = 0;
	m_nick = "Unknown";
	m_address = "Unknown";
	m_type = PT_UNKNOWN;
}

bool Person::QueryInfo()
{
	String2 url;
	url.Printf("http://www.friendster.com/user.jsp?id=%d", m_id);

	String2 str;
	ReqContext cl;
	cl.GetURL(url, str);
	if (str == "")
		return false;

	int offset;

	// first check if this is a suspended account
	offset = str.find("<div class=\"heading\">Account Suspended");
	if (offset != -1)
	{
		m_type = PT_SUSPENDED;
		return true;
	}

	// then grab nick
	offset = str.find("Friendster - ");
	if (offset == -1)
		return false;
	ExtractLine(str, offset+13, m_nick);

	// and address
	offset = str.find("Location:");
	if (offset != -1)
	{
		SkipToNextLine(str, offset);
		SkipToNextLine(str, offset);
		String2 addy;
		ExtractLine(str, offset, addy);
		addy.TrimLeft();
		addy.TrimRight();
		m_address = addy;
	}
	printf(" Person %d: '%s', '%s'\n", m_id, (const char *)m_nick, (const char *)m_address);

	return true;
}

bool Person::QueryFriends()
{
	String2 url;

	const char *pattern = "class=friend href=\"/user.jsp?id=";

	int page = 0;
	bool bMorePages = true;
	while (bMorePages)
	{
		url.Printf("http://www.friendster.com/friends.jsp?id=%d&start=%d", m_id, page*40);

		String2 str;
		ReqContext cl;
		cl.GetURL(url, str);
		if (str == "")
			return false;

//		printf(" Friends:");
		int on_this_page = 0;
		String2 out;
		int offset = 0;
		while (1)
		{
			offset = str.find(pattern, offset);
			if (offset == -1)
				break;
			offset += strlen(pattern);
			ExtractLine(str, offset, out);
			int id = atoi(out);
			m_friends.push_back(id);
			on_this_page++;
//			printf(" %d", id);
		}
//		printf("\n");
		if (on_this_page < 40)
			bMorePages = false;
		page++;
	}
	int num = m_friends.size();
	printf(" Person %d: %d friend%s\n", m_id, num, num != 1 ? "s" : "");

	return true;
}

class People
{
public:
	void AddPerson(Person *per);
	Person *GetNextPersonToQuery();
	Person *GetPerson(int id);
	void StartWithPerson(int home_person);
	void Gather(int num_people);

	bool Read(const char *fname);
	bool Write(const char *fname);
	bool WriteNeato(const char *fname, int maxhops);
	void PrintStats();

	int m_home_person;
	int m_tier;
	int m_tiers;

	std::vector<Person*> m_people;
	std::map<int, Person*> m_lookup;
};

void People::AddPerson(Person *per)
{
	m_people.push_back(per);
	m_lookup[per->m_id] = per;
}

Person *People::GetNextPersonToQuery()
{
	Person *per;

	while (m_tier < m_tiers)
	{
		unsigned int i, size = m_people.size();
		for (i = 0; i < size; i++)
		{
			per = m_people[i];
			if (per->m_hops == m_tier &&
				per->NumFriends() == 0 &&
				per->m_type != PT_SUSPENDED)
				return per;
		}
		m_tier++;
	}
	return NULL;
}

Person *People::GetPerson(int id)
{
	return m_lookup[id];
}

void People::StartWithPerson(int home_person)
{
	// create the first person
	Person *p = new Person();
	p->m_id = home_person;
	AddPerson(p);

	// start with a single tier consisting of the first person
	m_home_person = home_person;
	m_tiers = 1;
	m_tier = 0;
}

void People::Gather(int num_people)
{
	int done = 0;
	for (; num_people; num_people--)
	{
		printf("(%d done, %d left)\n", done, num_people);

		Person *p = GetNextPersonToQuery();
		if (!p)
			return;

		// Should be able to get this info; if not, e.g. if the server is
		//  being flaky or timing out, then just exit, since we can resume
		//  from this state later.
		if (!p->QueryInfo())
			return;

		if (p->m_type == PT_SUSPENDED)
			continue;

		if (!p->QueryFriends())
			return;

		// Add any new people we've found
		unsigned int i, size = p->m_friends.size();
		for (i = 0; i < size; i++)
		{
			int id = p->m_friends[i];
			if (!GetPerson(id))
			{
				Person *newp = new Person();
				newp->m_id = id;

				// We are encountering them for the first time in a breadth-first
				//  traversal, so we know this is the shortest number of hops.
				newp->m_hops = p->m_hops + 1;
				AddPerson(newp);

				if (newp->m_hops >= m_tiers)
					m_tiers = newp->m_hops + 1;
			}
		}
		done++;
	}
}

bool People::Write(const char *fname)
{
	FILE *fp = fopen(fname, "wb");
	if (!fp)
		return false;

	unsigned int i, j, size = m_people.size();
	fprintf(fp, "%d people\n", size);

	for (i = 0; i < size; i++)
	{
		Person *per = m_people[i];
		fprintf(fp, "id %d\n", per->m_id);
		fprintf(fp, "type %d\n", per->m_type);
		fprintf(fp, "%s\n", (const char *) per->m_nick);
		fprintf(fp, "%s\n", (const char *) per->m_address);
		fprintf(fp, "%d\n", per->m_hops);

		unsigned int friends = per->m_friends.size();
		fprintf(fp, "%d: ", friends);
		for (j = 0; j < friends; j++)
		{
			fprintf(fp, " %d", per->m_friends[j]);
		}
		fprintf(fp, "\n");
	}

	fclose(fp);
	return true;
}


// dumb but functional way of storing a bunch of connections
class conn {
public:
	conn(int i1, int i2) { id1 = i1; id2 = i2; }
	int id1, id2;
};
typedef std::vector<conn> CMap;


bool People::WriteNeato(const char *fname, int maxhops)
{
	FILE *fp = fopen(fname, "wb");
	if (!fp)
		return false;

	bool bHideLeaves = false;

	unsigned int i, j, size = m_people.size(), num_close = 0;
	int id1, id2;
	Person *per1, *per2;

	// build array of unique connections ("friendships")
	printf("  Building array of unique connections (\"friendships\") ...");
	CMap cmap;
	for (i = 0; i < size; i++)
	{
		per1 = m_people[i];
		for (j = 0; j < per1->m_friends.size(); j++)
		{
			id1 = per1->m_id;
			id2 = per1->m_friends[j];
			// sort into ascending order to avoid backlinks
			if (id2 < id1)
			{
				int temp = id1;
				id1 = id2;
				id2 = temp;
			}
			// check (dumb slow way) for uniqueness
			bool bFound = false;
			for (CMap::iterator it = cmap.begin(); it != cmap.end(); it++)
			{
				if (it->id1 == id1 && it->id2 == id2)
				{
					bFound = true;
					break;
				}
			}
			if (!bFound)
			{
				conn newconn(id1,id2);
				cmap.push_back(newconn);
			}
		}
	}
	for (CMap::iterator it = cmap.begin(); it != cmap.end(); it++)
	{
		id1 = it->id1;
		id2 = it->id2;
		per1 = GetPerson(id1);
		per2 = GetPerson(id2);

		if (!per1 || !per2)	// safety check
			continue;	// something went wrong

		if (per1->m_hops > maxhops || per2->m_hops > maxhops)
			continue;

		num_close++;
	}

	printf("\n  Found %d friendships, %d within %d hops, writing file ...",
		cmap.size(), num_close, maxhops);

	fprintf(fp, "graph FRIENDSTER {\n");
	fprintf(fp, "\tnode [shape=box,height=.1,width=.2,fontname=Arial,fontsize=8];\n");

	for (CMap::iterator it = cmap.begin(); it != cmap.end(); it++)
	{
		id1 = it->id1;
		id2 = it->id2;
		per1 = GetPerson(id1);
		per2 = GetPerson(id2);

		if (!per1 || !per2)	// safety check
			continue;	// something went wrong

		if (bHideLeaves)
		{
			if (per1->NumFriends() == 1 || per2->NumFriends() == 1)
				continue;
		}
		if (per1->m_hops > maxhops || per2->m_hops > maxhops)
			continue;

		// We need to use the id somehow, because there are plenty of duplicate
		// nicknames, we have to tell Graphviz that they are distinct labels.
//		fprintf(fp, "\t%d -- %d\n", id1, id2);
		fprintf(fp, "\t\"%s\\n%d\" -- \"%s\\n%d\"",
			(const char *)per1->m_nick, id1,
			(const char *)per2->m_nick, id2);
		if (id1 == m_home_person || id2 == m_home_person)
			fprintf(fp, " [len=2]");
		fprintf(fp, ";\n");
	}
	fprintf(fp, "}\n\n");

	fclose(fp);
	return true;
}

void get_line_from_file(FILE *fp, string &str)
{
	char buf[900];
	fgets(buf, 900, fp);
	int len = strlen(buf);
	while (buf[len-1] == '\r' || buf[len-1] == '\n')
	{
		buf[len-1] = 0;
		len--;
	}
	str = buf;
}

bool People::Read(const char *fname)
{
	FILE *fp = fopen(fname, "rb");
	if (!fp)
		return false;

	m_tier = 0;
	m_tiers = 1;

	unsigned int i, j, num, id;
	fscanf(fp, "%d people\n", &num);
	for (i = 0; i < num; i++)
	{
		Person *per = new Person();

		fscanf(fp, "id %d\n", &per->m_id);
		fscanf(fp, "type %d\n", &per->m_type);
		get_line_from_file(fp, per->m_nick);
		get_line_from_file(fp, per->m_address);
		fscanf(fp, "%d\n", &per->m_hops);

		if (per->m_hops >= m_tiers)
			m_tiers = per->m_hops+1;

		unsigned int friends;
		fscanf(fp, "%d: ", &friends);
		for (j = 0; j < friends; j++)
		{
			fscanf(fp, " %d", &id);
			per->m_friends.push_back(id);
		}
		fscanf(fp, "\n");
		AddPerson(per);
	}

	// start with a single tier consisting of the first person
	if (num > 1)
		m_home_person = m_people[0]->m_id;

	fclose(fp);
	return true;
}

void People::PrintStats()
{
	unsigned int i, size = m_people.size(), known = 0;

	for (i = 0; i < size; i++)
	{
		Person *per = m_people[i];
		if (per->m_nick != "Unknown")
			known++;
	}
	printf("stats: There are %d people.\n", size);
	printf("       There are details known for %d of them.\n", known);
}

void main(int argc, char **argv)
{
	People people;
	int i;
	bool success;

	bool bShowHelp = false;
	if (argc < 2)
		bShowHelp = true;

	for (i = 1; i < argc; i++)
	{
		if (!strcmp(argv[i], "-email"))
		{
			// Need these cookies, in order to do any queries at all
			AddCookie("friendster_email", argv[i+1]);
			i++;
		}
		else if (!strcmp(argv[i], "-password"))
		{
			// Need these cookies, in order to do any queries at all
			AddCookie("friendster_password", argv[i+1]);
			i++;
		}
		else if (!strcmp(argv[i], "-read"))
		{
			printf("reading file '%s'...", argv[i+1]);
			success = people.Read(argv[i+1]);
			if (success)
				printf(" success.\n");
			else
				printf(" failed.\n");
			i++;
			people.PrintStats();
		}
		else if (!strcmp(argv[i], "-start"))
		{
			int start_id = atoi(argv[i+1]);
			printf("starting with person %d\n", start_id);
			people.StartWithPerson(start_id);
			i++;
		}
		else if (!strcmp(argv[i], "-gather"))
		{
			int number = atoi(argv[i+1]);
			printf("gathering information on %d people\n", number);
			people.Gather(number);
			i++;
		}
		else if (!strcmp(argv[i], "-write"))
		{
			printf("writing file '%s'...", argv[i+1]);
			success = people.Write(argv[i+1]);
			if (success)
				printf(" success.\n");
			else
				printf(" failed.\n");
			i++;
		}
		else if (!strcmp(argv[i], "-neato"))
		{
			printf("writing neato file '%s'...\n", argv[i+1]);
			success = people.WriteNeato(argv[i+1], atoi(argv[i+2]));
			if (success)
				printf(" success.\n");
			else
				printf(" failed.\n");
			i+=2;
		}
		else
		{
			// printf("unknown argument '%s'\n);
			bShowHelp = true;
		}
	}
	if (bShowHelp)
	{
		printf("\nfrspider - spider for the Friendster website\n\n");
		printf("Usage - Command-line arguments should be providing in this order:\n");
		printf("   -email <address>  Provide the address that Friendster knows you by.\n");
		printf("   -password <pass>  Provide the password for your Friendster account.\n");
		printf("   -read <file>      Read an existing frspider data file.\n");
		printf("   -start <id>       The ID of a user (generally your own) from which to\n"
			   "                      start gathering data.\n");
		printf("   -gather <num>     Gather information on a given number of people.\n");
		printf("   -write <file>     Write to a frspider data file.\n");
		printf("   -neato <file> <maxhops>\n"
			   "       Write to a file that can be passed to Graphviz neato.\n");
	}
}

